/*
 * Copyright (C) 2001 Igor Sysoev.
 */

#include "mod_accel.h"

#ifdef MOD_ACCEL_CACHEMGR

static char *accel_mgr_status[] =
{
    "success", "invalid", "no_accelerated", "not_found",
    "no_cachable", "cache_error", "backend_error"
};

int accel_cachemgr(request_rec *r)
{
    int             action = ACCEL_MGR_NO_ACTION, status, rc;
    char           *url = NULL, *result = NULL, *result_perfect,
                   *error = NULL, *reason = NULL;
    accel_rec      *a = NULL;
    request_rec    *sr = NULL;

    if (r->args && strncmp(r->args, "action=", 7) == 0) {
        if (strncmp(&r->args[7], "refresh&", 8) == 0) {
            action = ACCEL_MGR_REFRESH;
            url = &r->args[15];

        } else if (strncmp(&r->args[7], "remove&", 7) == 0) {
            action = ACCEL_MGR_REMOVE;
            url = &r->args[14];

        } else {
            status = ACCEL_MGR_INVALID;
        }

    } else {
        status = ACCEL_MGR_INVALID;
    }

    if (url && strncmp(url, "url=", 4) == 0) {
        url += 4;
    } else {
        status = ACCEL_MGR_INVALID;
    }

    if (action) {
        sr = ap_sub_req_lookup_uri(url, r);
        if (sr->proxyreq == ACCEL) {

            if (!(a = ap_ctx_get(sr->ctx, "mod_accel"))) {
                if (strncmp(sr->filename, "proxy:", 6) == 0) {
                    a = accel_init_ctx(sr);
                    a->url = accel_make_url(sr, sr->filename + 6);
                    a->pass = ap_pcalloc(sr->pool, sizeof(accel_pass));
                }
            }
        }
    }

    if (action == ACCEL_MGR_REMOVE) {
        if (a) {
            a->dir_conf = (accel_dir_conf *)
                        ap_get_module_config(sr->per_dir_config, &accel_module);
            accel_init_dir_config(a, a->dir_conf);

            a->url = ap_pstrcat(r->pool, a->url,
                                ap_os_escape_path(r->pool, url + 1, 1),
                                NULL);

            accel_make_cachefile_name(a, 1);
            a->cachefile_exists = 1;
            if (accel_unlink_cachefile(a) == -1) {
                if (errno == ENOENT) {
                    status = ACCEL_MGR_NOT_FOUND;
                    error = "Cache file was not found";
                } else {
                    status = ACCEL_MGR_CACHE_ERROR;
                    error = "Cache error was encountered";
                }
            } else {
                status = ACCEL_MGR_SUCCESS;
                a->accel_state = ACCEL_RMVD;
            }

            result = "remove";
            result_perfect = "removed";
            accel_set_notes(a);

        } else {
            status = ACCEL_MGR_NO_ACCELERATED;
            reason = "no accelerated";
        }

    } else if (action == ACCEL_MGR_REFRESH) {
        if (a) {
            a->cachemgr = 1;
            a->invalidate = 1;
            a->cache_body = 1;

            rc = ap_run_sub_req(sr);

            if (a->cachefile_existed > 0) {
                result = "refresh";
                result_perfect = "refreshed";
            } else {
                result = "create";
                result_perfect = "created";
            }

            if (rc == OK && sr->status == HTTP_OK) {
                if (a->expired) {
                    status = ACCEL_MGR_NO_CACHABLE;
                    reason = "no cachable";
                } else {
                    status = ACCEL_MGR_SUCCESS;
                }

            } else if (rc == OK && sr->status == HTTP_NOT_FOUND) {
                status = ACCEL_MGR_NOT_FOUND;
                error = "Backend response with 404 error";

            } else if (a->cache_error) {
                status = ACCEL_MGR_CACHE_ERROR; 
                error = "Cache error was encountered";

            } else if (a->backend_aborted
                     || a->backend_timedout
                     || ap_is_HTTP_SERVER_ERROR(rc)) {
                status = ACCEL_MGR_BACKEND_ERROR; 
                error = "Backend error was encountered";
            }

        } else {
            status = ACCEL_MGR_NO_ACCELERATED;
            reason = "no accelerated";
        }
    }

    if (sr)
        ap_destroy_sub_req(sr);

    if (result)
        ap_table_setn(r->headers_out, "X-Accel-Cachemgr-Action", result);

    if (url)
        ap_table_setn(r->headers_out, "X-Accel-Cachemgr-URL", url);

    ap_table_setn(r->headers_out, "X-Accel-Cachemgr-Status",
                  accel_mgr_status[status]);

    r->content_type = "text/html";
    ap_send_http_header(r);

    ap_soft_timeout("accel: send cachemgr body", r);
    ap_rvputs(r,
              DOCTYPE_HTML_2_0
              "<HTML><HEAD>\n<TITLE>" "mod_accel cache manager"
              "</TITLE>\n</HEAD><BODY>\n<H1>" "mod_accel cache manager"
              "</H1>\n",
              NULL);

    if (status == ACCEL_MGR_INVALID)
        ap_rvputs(r,
                  "Invalid aruments for accel-cachemgr",
                  NULL);

    else if (reason)
        ap_rvputs(r,
                  "URI <A HREF=\"", url, "\">", url, "</A> is ", reason,
                  NULL);

    else if (error)
        ap_rvputs(r,
                  error, " while trying to ", result,
                  " cache file for URI <A HREF=\"", url, "\">", url, "</A>",
                  NULL);

    else
        ap_rvputs(r,
                  "Cache file for URI <A HREF=\"", url, "\">", url,
                  "</A> is ", result_perfect, NULL);

    ap_rputs(ap_psignature("<HR>\n", r), r);
    ap_rputs("</BODY></HTML>\n", r);
    ap_kill_timeout(r);

    return OK;
}
#endif
