/*
 * Copyright (C) 2001 Igor Sysoev.
 */

#include "mod_accel.h"

#ifdef EAPI_MM

int accel_check_busy_lock(accel_rec *a, int busy_lock, int wait)
{
    int i;

    int busy = FALSE;
    int connected = 0;
    int waiting = 0;

    if (busy_lock <= 0)
        return FALSE;

    /* it seems that the best place to get child's scoreboard slot number
       is child init phase but sometimes this number is not set at child init
       phase so here is the reliable place to get it. At least in 1.3.14 */
    if (accel_slot_num == -1)
        accel_set_slot_num();

#ifdef MOD_ACCEL_DEBUG
    ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, a->r,
                  "accel: lock %d", busy_lock);
    ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, a->r,
                  "accel: slot %d: %d %s", accel_slot_num, a->time, a->md5);
#endif

    ap_acquire_pool(shared_pool, AP_POOL_RW);

    for (i = 0; i < ap_daemons_limit; i++) {
        if (i == accel_slot_num) {
            accel_state[i].state = ACCEL_FREE;
            continue;
        }

#ifdef MOD_ACCEL_DEBUG
        if (accel_state[i].time == 0)
            continue;

        ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, a->r,
                      "accel: slot %d: %d %d %s",
                      i, accel_state[i].state,
                      a->time - accel_state[i].time, accel_state[i].md5);
#endif

        if (accel_state[i].state == ACCEL_FREE)
            continue;

        if (accel_state[i].state == ACCEL_BUSY_LOCK
            && a->time - accel_state[i].time <= busy_lock
            && strncmp(accel_state[i].md5, a->md5, MD5_LEN) == 0)
        {
            busy = TRUE;
        }

        if (a->time - accel_state[i].time < TOO_LONG_TIME
            && a->pass->sum == accel_state[i].sum) {

            if (accel_state[i].state == ACCEL_BUSY_LOCK
                || accel_state[i].state == ACCEL_BUSY_PASS)
            {
                connected++;

            } else if (accel_state[i].state == ACCEL_WAIT) {

                waiting++;
            }
        }
    }

    a->max_connected_checked = 1;
    a->max_connected = 0;
    a->max_waiting = 0;

    if (a->pass->max_conn) {

        if (a->pass->max_conn <= connected)
            a->max_connected = 1;

        if (a->pass->max_wait <= waiting)
            a->max_waiting = 1;
    }

    if (accel_slot_num != -1) {

        if (!busy && !a->max_connected) {
            accel_state[accel_slot_num].state = ACCEL_BUSY_LOCK;
            accel_state[accel_slot_num].time = a->time;
            accel_state[accel_slot_num].sum = a->pass->sum;
            ap_cpystrn(accel_state[accel_slot_num].md5, a->md5, MD5_LEN + 1);

            a->busy_lock = 1;

        } else if (wait && !a->max_waiting) {
            accel_state[accel_slot_num].state = ACCEL_WAIT;
            accel_state[accel_slot_num].time = a->time;
            accel_state[accel_slot_num].sum = a->pass->sum;

            a->busy_lock = 1;
        }
    }

    ap_release_pool(shared_pool);

    return busy;
}

void accel_clear_busy_lock(accel_rec *a)
{
    if (a->busy_lock) {
        ap_acquire_pool(shared_pool, AP_POOL_RW);
        accel_state[accel_slot_num].state = ACCEL_FREE;
        ap_release_pool(shared_pool);

        a->busy_lock = 0;
   }
}

int accel_max_connected(accel_rec *a)
{
    int i;
    int connected = 0;
    int waiting = 0;

    if (!a->pass->max_conn) {
        a->max_connected_checked = 1;
        a->max_connected = 0;
        a->max_waiting = 0;
        return FALSE;
    }

    if (a->max_connected_checked)
       return a->max_connected;

    /* see comment above with the same code in accel_check_busy_lock() */
    if (accel_slot_num == -1)
        accel_set_slot_num();

    ap_acquire_pool(shared_pool, AP_POOL_RW);

    for (i = 0; i < ap_daemons_limit; i++) {
        if (i == accel_slot_num) {
            accel_state[i].state = ACCEL_FREE;
            continue;
        }

#ifdef MOD_ACCEL_DEBUG
        if (accel_state[i].time == 0)
            continue;

        ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, a->r,
                      "accel: slot %d: %d %d %x",
                      i, accel_state[i].state,
                      a->time - accel_state[i].time, accel_state[i].sum);
#endif

        if (accel_state[i].state == ACCEL_FREE)
            continue;

        if (a->time - accel_state[i].time < TOO_LONG_TIME
            && a->pass->sum == accel_state[i].sum) {

            if (accel_state[i].state == ACCEL_BUSY_LOCK
                || accel_state[i].state == ACCEL_BUSY_PASS)
            {
                connected++;

            } else if (accel_state[i].state == ACCEL_WAIT) {

                waiting++;
            }
        }
    }

    a->max_connected_checked = 1;
    a->max_connected = 0;
    a->max_waiting = 0;

    if (a->pass->max_conn) {

        if (a->pass->max_conn <= connected)
            a->max_connected = 1;

        if (a->pass->max_wait <= waiting)
            a->max_waiting = 1;
    }

    if (accel_slot_num != -1 && !a->max_connected) {
        accel_state[accel_slot_num].state = ACCEL_BUSY_PASS;
        accel_state[accel_slot_num].time = a->time;
        accel_state[accel_slot_num].sum = a->pass->sum;

        a->busy_lock = 1;
    }

    ap_release_pool(shared_pool);

    return a->max_connected;
}

void accel_set_slot_num(void)
{
    int i, pid;
             
    pid = getpid();
    
    ap_sync_scoreboard_image();
    for (i = 0; i < ap_daemons_limit; i++) {
        if (pid == ap_scoreboard_image->parent[i].pid) {
            accel_slot_num = i;
            break;
        }
    }
}
        
#else  /* ndef EAPI_MM */

int accel_check_busy_lock(accel_rec *a, int busy_lock, int wait)
{
    return FALSE;
}

void accel_clear_busy_lock(accel_rec *a)
{
}

int accel_max_connected(accel_rec *a)
{
    return FALSE;
}

#endif
